from typing import Dict, List

import rdflib
from octadocs.types import QueryResult, SelectResult
from rdflib import term
from rdflib.plugins.sparql.processor import SPARQLResult


def query(
    query_text: str,
    instance: rdflib.ConjunctiveGraph,
    **kwargs: str,
) -> QueryResult:
    """Run SPARQL SELECT query and return formatted result."""
    sparql_result: SPARQLResult = instance.query(
        query_text,
        initBindings=kwargs,
    )

    if sparql_result.askAnswer is not None:
        return sparql_result.askAnswer

    if sparql_result.graph is not None:
        graph: rdflib.Graph = sparql_result.graph
        for prefix, namespace in instance.namespaces():
            graph.bind(prefix, namespace)

        return graph

    return _format_query_bindings(sparql_result.bindings)


def _format_query_bindings(
    bindings: List[Dict[rdflib.Variable, term.Identifier]],
) -> SelectResult:
    """
    Format bindings before returning them.

    Converts Variable to str for ease of addressing.
    """
    return SelectResult(
        {
            str(variable_name): rdf_value
            for variable_name, rdf_value
            in row.items()
        }
        for row in bindings
    )



from typing import NamedTuple, Union

from mkdocs.structure.nav import Link, Navigation, Section
from mkdocs.structure.pages import Page

NavigationItem = Union[Page, Section, Link, Navigation]
PAGE_DEFAULT_POSITION = 0


class SortKey(NamedTuple):
    """Sort key."""

    is_index: bool
    position: str
    title: str



from functools import cached_property
from typing import Optional, cast

from iolanta.facets.base import FacetSearchAttempt
from more_itertools import first
from rdflib.term import Literal, URIRef


class FindFacetByLiteralDatatype(FacetSearchAttempt):
    """
    Find facet for a literal value by data type.

    Look for such `?facet` that `{self.node.datatype}` `iolanta:datatypeFacet`
    `?facet`, whereas `?facet` `iolanta:supports` `{self.environment}`.
    """

    @cached_property
    def facet(self) -> Optional[URIRef]:
        """Find facet."""
        if not isinstance(self.node, Literal):
            return None

        if (datatype := self.node.datatype) is None:
            return None

        rows = self.ldflex.query(
            '''
            SELECT ?facet WHERE {
                $datatype iolanta:datatypeFacet ?facet .
                ?facet iolanta:supports $env .
            }
            ''',
            datatype=datatype,
            env=self.environment,
        )

        try:
            return cast(URIRef, first(rows)['facet'])
        except ValueError:
            return None



from dataclasses import dataclass
from typing import List

from documented import DocumentedError
from rdflib import URIRef
from rdflib.term import Node


@dataclass
class PageNotFound(DocumentedError):
    """
    Page not found by IRI: `{self.iri}`.

    !!! error "Page not found by IRI: `{self.iri}`"
        Every page on your documentation site, just as any other entity described on
        it, has an IRI — a unique identifier similar to a URL. IRI can be generated
        in two alternative ways.

          1. If you wanted to use IRI generated automatically, then please confirm
             that in the `docs` directory there is a file under the following path:

             ```
             {self.possible_path}
             ```

          2. If you wanted to use IRI that you specified explicitly then confirm
             there is a Markdown file somewhere in `docs` directory which contains
             the following text in its header:

             ```markdown
             ---
             $id: {self.possible_id}
             ---
             ```
    """

    iri: URIRef

    @property
    def possible_path(self) -> str:
        """Guess on a possible path."""
        base_path = str(self.iri).replace('local:', '')
        return f'{base_path}.md'

    @property
    def possible_id(self) -> str:
        """Guess on a possible $id."""
        return str(self.iri).replace('local:', '')


@dataclass
class MultipleFacetsFoundForTypes(DocumentedError):
    """
    Multiple suitable facets found.

    Unable to render the node unambiguously.

      - Node: {self.node}
      - Types:
    {self.formatted_types}
      - Facets found:
    {self.formatted_facets}
    """

    node: Node
    types: List[URIRef]
    facets: List[URIRef]

    @property
    def formatted_facets(self):
        return '\n'.join([
            f'    - {facet}'
            for facet in self.facets
        ])

    @property
    def formatted_types(self):
        return '\n'.join([
            f'    - <{node_type}>'
            for node_type in self.types
        ])




import sys
from dataclasses import dataclass
from typing import Dict, List, Optional, cast

import rdflib
# noinspection PyProtectedMember
from mkdocs.structure.nav import (  # noqa: WPS450
    Navigation,
    Section,
    _add_previous_and_next_links,
)
from mkdocs.structure.pages import Page
from octadocs.conversions import iri_by_page
from octadocs.navigation.nav_as_pages import create_pages_list_by_navigation
from octadocs.navigation.types import (
    PAGE_DEFAULT_POSITION,
    NavigationItem,
    SortKey,
)
from octadocs.query import query
from octadocs.types import SelectResult
from singledispatchmethod import singledispatchmethod

if sys.version_info >= (3, 8):
    from functools import cached_property  # noqa
else:
    from backports.cached_property import cached_property  # noqa: WPS433,WPS440


def is_index_md(navigation_item: NavigationItem) -> bool:
    """Determine if certain navigation item is an index.md page."""
    return (
        isinstance(navigation_item, Page) and
        navigation_item.file.src_path.endswith('index.md')
    )


def sort_key(navigation_item: NavigationItem) -> SortKey:
    """Determine sort key for a navigation item."""
    title = navigation_item.title
    if title is None:
        try:
            title = navigation_item.file.name
        except AttributeError:
            title = ''

    return SortKey(
        is_index=not is_index_md(navigation_item),
        position=str(navigation_item.position),
        title=title,
    )


def find_index_page_in_section(section: Section) -> Optional[Page]:
    """Find index.md page in the section, if any."""
    index_pages = list(filter(
        is_index_md,
        section.children,
    ))

    if index_pages:
        return index_pages[0]

    return None


@dataclass
class PageFromGraph:
    """Information about a Page retrieved from Octiron graph."""

    title: Optional[str] = None
    position: int = PAGE_DEFAULT_POSITION


@dataclass(repr=False)
class OctadocsNavigationProcessor:
    """Rewrite navigation based on the knowledge graph."""

    graph: rdflib.ConjunctiveGraph
    navigation: Navigation

    @cached_property
    def pages_from_graph(self) -> Dict[rdflib.URIRef, PageFromGraph]:
        """Fetch the position by page from graph."""
        query_text = '''
            SELECT ?page ?position ?title WHERE {
                ?page a octa:Page .
                ?thing octa:subjectOf ?page .

                OPTIONAL {
                    ?thing octa:position ?position .
                }

                OPTIONAL {
                    ?thing octa:title ?title .
                }
            }
        '''

        rows = query(
            query_text=query_text,
            instance=self.graph,
        )

        return {
            row['page']: PageFromGraph(
                title=getattr(
                    row.get('title'),
                    'value',
                    None,
                ),
                position=getattr(
                    row.get('position'),
                    'value',
                    PAGE_DEFAULT_POSITION,
                ),
            )
            for row in cast(SelectResult, rows)
        }

    def generate(self) -> Navigation:
        """Generate the navigation structure."""
        # Rearrange navigation itself
        navigation = self.rearrange_navigation(self.navigation)

        # Generate flat pages list
        pages = list(create_pages_list_by_navigation(navigation))
        navigation.pages = pages
        _add_previous_and_next_links(navigation.pages)

        return navigation

    @singledispatchmethod
    def rearrange_navigation(
        self,
        navigation_item: NavigationItem,
    ) -> NavigationItem:
        """Change the order of navigation menu elements."""
        raise NotImplementedError()

    @rearrange_navigation.register
    def _rearrange_page(self, page: Page) -> Page:
        iri = iri_by_page(page)
        page_from_graph = self.pages_from_graph.get(iri)

        if page_from_graph:
            page.title = page_from_graph.title
            page.position = page_from_graph.position

        else:
            page.position = PAGE_DEFAULT_POSITION

        return page

    def _rearrange_list_of_navigation_items(
        self,
        navigation_items: List[NavigationItem],
    ) -> List[NavigationItem]:
        navigation_items = list(map(
            self.rearrange_navigation,
            navigation_items,
        ))

        return sorted(
            navigation_items,
            key=sort_key,
        )

    @rearrange_navigation.register
    def _rearrange_section(self, section: Section) -> Section:
        section.children = self._rearrange_list_of_navigation_items(
            section.children,
        )

        index_page = find_index_page_in_section(section)
        if index_page is None:
            section.position = PAGE_DEFAULT_POSITION

        else:
            section.position = index_page.position

            if index_page.title:
                section.title = index_page.title

        return section

    @rearrange_navigation.register
    def _rearrange_navigation(self, navigation: Navigation):
        navigation.items = self._rearrange_list_of_navigation_items(
            navigation.items,
        )

        return navigation



from dataclasses import dataclass
from typing import Optional

from documented import DocumentedError
from rdflib import URIRef


@dataclass
class YAMLError(DocumentedError):
    """
    Invalid YAML.

    File: {self.iri}

    {self.error}
    """

    iri: Optional[URIRef]
    error: Exception



import html
import textwrap

import funcy
from dominate.tags import br

from iolanta.facet import Facet
from octadocs_ibis.mixins import LanguageAware


class IBISNode(LanguageAware, Facet):
    sparql_query = '''
    SELECT * WHERE {
        $iri
            octa:title ?title ;
            a ?type .

        OPTIONAL {
            $iri octa:symbol ?symbol .
        }

        ?type rdfs:label ?type_title .
    } ORDER BY DESC(LANG(?type_title))
    '''

    def render(self):
        rows = self.query(
            self.sparql_query,
            iri=self.uriref,
        )

        row = funcy.first(rows)

        if row is None:
            return str(self.uriref)

        heading = row['title']
        if symbol := row.get('symbol'):
            heading = f'{symbol} {heading}'

        wrapped_title = textwrap.wrap(
            html.escape(heading.replace('"', '')),
            width=20,
        )

        wrapped_title_with_line_breaks = funcy.interpose(
            # See:
            #   https://renenyffenegger.ch/notes/tools/Graphviz/attributes/
            #   label/HTML-like/index
            br().render(xhtml=True),
            wrapped_title,
        )

        return ''.join(wrapped_title_with_line_breaks)



from dominate.tags import a
from more_itertools import first

from iolanta.facet import Facet
from octadocs.iolanta import render


class ProvEntity(Facet):
    """Render a node with prov:value defined."""

    def render(self):
        """Render value and link to it."""
        rows = self.query(
            '''
            SELECT * WHERE {
                $node prov:value ?value .

                OPTIONAL {
                    $node prov:wasDerivedFrom ?source .
                }

                OPTIONAL {
                    $node rdfs:comment ?comment .
                }
            }
            ''',
            node=self.iri,
        )

        try:
            row = first(rows)
        except ValueError as err:
            raise ValueError(
                f'No `prov:value` was found for the node: {self.uriref}.'
            ) from err

        kwargs = {}
        if comment := row.get('comment'):
            kwargs.update(title=comment)

        rendered_value = render(
            node=row['value'],
            octiron=self.octiron,
        )

        if row.get('source'):
            return a(
                rendered_value,
                href=row['source'],
                target='_blank',
                **kwargs,
            )

        return rendered_value



from dataclasses import dataclass
from functools import cached_property
from typing import Optional, Union

from ldflex import LDFlex
from octadocs.octiron import Octiron
from rdflib.term import Node, URIRef, BNode


@dataclass
class Facet:
    """Base facet class."""

    iri: Node
    octiron: Octiron
    environment: Optional[URIRef] = None

    @property
    def ldflex(self) -> LDFlex:
        """Extract LDFLex instance."""
        return self.octiron.ldflex

    @cached_property
    def uriref(self) -> Union[URIRef, BNode]:
        """Format as URIRef."""
        if isinstance(self.iri, BNode):
            return self.iri

        return URIRef(self.iri)

    def query(self, query_text: str, **kwargs):
        """SPARQL query."""
        return self.ldflex.query(
            query_text=query_text,
            **kwargs,
        )

    def render(self):
        """Render the facet."""
        raise NotImplementedError()

    @property
    def language(self):
        return self.octiron.language

    def __str__(self):
        """Render."""
        return str(self.render())



import operator

from dominate.tags import img
from more_itertools import first

from iolanta.facet import Facet


class Image(Facet):
    sparql_query = 'SELECT * WHERE { $image octa:url ?url } LIMIT 1'

    def render(self):
        rows = self.query(
            self.sparql_query,
            image=self.iri,
        )

        urls = map(
            operator.itemgetter('url'),
            rows,
        )

        try:
            url = first(urls)
        except ValueError:
            return f'(No Image for {self.uriref} [{type(self.uriref)}])'

        return img(src=url)



from dominate.tags import a, code, table, tbody, td, th, thead, tr
from octadocs.octiron import Octiron
from rdflib import URIRef


def build_table_row(status) -> tr:
    raw = str(status['status']).replace('local:', '')

    label = status['label']

    if symbol := status.get('symbol'):
        label = f'{symbol} {label}'

    defined_by_label = status.get('defined_by_label')

    if defined_by_url := status.get('defined_by_url'):
        defined_by = a(
            defined_by_label or defined_by_url,
            href=defined_by_url,
            target='_blank',
        )

    elif defined_by_label:
        defined_by = defined_by_label

    else:
        defined_by = code(status['defined_by_iri'])

    return tr(
        td(code(raw)),
        td(label),
        td(defined_by or '')
    )


def status_class(octiron: Octiron, iri: URIRef):
    """Visualize all available status values as a table."""
    choices = octiron.query(
        '''
        SELECT
            ?status ?label ?symbol
            ?defined_by_iri ?defined_by_url ?defined_by_label
        WHERE {
            ?status a adr:Status .

            GRAPH ?defined_by_iri {
                ?status rdfs:label ?label .

                OPTIONAL {
                   ?status octa:symbol ?symbol .
                }
            }

            OPTIONAL {
                ?defined_by_iri octa:url ?defined_by_url .
            }

            OPTIONAL {
                ?defined_by_iri rdfs:label ?defined_by_label .
            }
        } ORDER BY ?label
        '''
    )

    rows = map(build_table_row, choices)

    return table(
        thead(
            tr(
                th('Code'),
                th('Label'),
                th('Defined By'),
            )
        ),
        tbody(*rows)
    )



from dominate.tags import blockquote, div, p
from more_itertools import first

from iolanta.facet import Facet
from iolanta.namespaces import IOLANTA
from octadocs.iolanta import render


class SourceForPosition(Facet):
    sparql_query = '''
    SELECT ?comment WHERE {
        $source rdfs:comment ?comment .
    } LIMIT 1
    '''

    def render(self):
        rows = self.query(
            self.sparql_query,
            source=self.iri,
        )

        try:
            row = first(rows)
        except ValueError:
            return f'[{self.uriref}]'

        return blockquote(
            p(
                row['comment'],
            ),
            p(
                render(
                    self.iri,
                    environments=[IOLANTA.html],
                    octiron=self.octiron,
                ),
                style='text-align: right',
            ),
        )



import json
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Optional, cast, Iterable, List

import funcy

from iolanta.facets.base import FacetSearchAttempt
from more_itertools import first
from rdflib.term import URIRef

from iolanta.facets.errors import MultipleFacetsFoundForTypes


QUERY = (
    Path(__file__).parent / 'sparql/find_facet_by_instance_types.sparql'
).read_text()


@dataclass
class FindFacetByInstanceTypes(FacetSearchAttempt):
    """Find facet by node type."""

    @cached_property
    def facet(self) -> Optional[URIRef]:
        """Find facet."""
        rows = self.ldflex.query(
            QUERY,
            env=self.environment,
            node=self.node,
        )

        facets = list(set(funcy.pluck('facet', rows)))

        if len(facets) > 1:
            raise MultipleFacetsFoundForTypes(
                node=self.node,
                types=list(funcy.pluck('type', rows)),
                facets=facets,
            )

        return first(facets, None)



import operator
from typing import Optional, Iterable, Tuple, Union

import pygraphviz
from funcy import distinct
from rdflib import Literal

from iolanta.facet import Facet
from ldflex.ldflex import SelectRow
from octadocs.iolanta import render
from octadocs_ibis.mixins import LanguageAware
from octadocs_ibis.models import IBIS


class IBISGraph(LanguageAware, Facet):
    """Draw graph."""

    nodes_query = '''
    SELECT ?node ?node_type ?title ?url ?is_accepted WHERE {
        $this octa:subjectOf ?current_page .

        ?current_directory octa:isParentOf ?current_page .

        {
            ?current_directory octa:isParentOf ?page .
        } UNION {
            ?current_directory octa:isParentOf / octa:isParentOf ?page .
            ?page a octa:IndexPage .
        }

        GRAPH ?page {
            ?node octa:title ?title .
        }

        ?node a skos:Concept, ?node_type .

        OPTIONAL {
            ?node octa:url ?url .
        }

        ?node_type octa:position ?node_type_priority .

        OPTIONAL {
            ?node ibis:evaluation / prov:value ?is_accepted .
            FILTER(!isBlank(?is_accepted))
        }

        FILTER(?node_type IN (ibis:Issue, ibis:Position, ibis:Argument))
    } ORDER BY ?node ?node_type_priority
    '''

    edges_query = '''
    SELECT ?edge ?source ?destination ?edge_label WHERE {
        ?source ?edge ?destination .
        ?edge rdfs:label ?edge_label .

        FILTER(
            ?edge IN (
                ibis:suggests,
                ibis:response,
                ibis:supported-by,
                ibis:opposed-by
            )
        )
    } ORDER BY DESC(LANG(?edge_label))
    '''

    def _node_options(self, row: SelectRow):
        if not row.get('url'):
            return {}

        is_accepted_literal: Optional[Literal] = row.get('is_accepted')

        if is_accepted_literal is None:
            # This is not an ibis:Position, or it does not have an evaluation.
            return {'fillcolor': '#CCCCCC', 'color': '#00000052'}

        is_accepted: bool = is_accepted_literal.toPython()

        if is_accepted:
            return {'fillcolor': '#00c853bb', 'color': '#00c853'}
        else:
            return {'fillcolor': '#ff5252bb', 'color': '#ff5252'}

    def _edge_options(
        self, row: SelectRow,
    ) -> Iterable[Tuple[str, Optional[Union[str, int]]]]:
        """Construct  edge options depending on its type."""
        if row['edge'] == IBIS.term('supported-by'):
            yield 'tooltip', '➕ За'

    def distinct_nodes(self, rows):
        return distinct(
            rows,
            key=operator.itemgetter('node'),
        )

    def _construct_node_rows(self):
        yield from self.distinct_nodes(
            self.query(self.nodes_query, this=self.uriref),
        )

    def render(self):
        node_rows = list(self._construct_node_rows())

        graph = pygraphviz.AGraph(
            directed=True,
            rankdir='LR',
            bgcolor='#ffffff00',
            splines='ortho',
        )

        for row in node_rows:
            rendered_node = render(
                row['node'],
                environments=[IBIS.IBISNode],
                octiron=self.octiron,
            )

            node_style = 'rounded,filled'

            node_shape = 'box'
            pen_width = 1
            if row['node'] == self.uriref:
                node_style = f'{node_style},bold'
                pen_width = 4

            if href := row.get('url'):
                href_options = {'href': href}
                if not href.startswith('/'):
                    # This is an external link.
                    rendered_node = f'{rendered_node} ⧉'
                    href_options.update(target='_blank')

            else:
                node_style = 'rounded,filled'
                href_options = {}

            graph.add_node(
                row['node'],
                shape=node_shape,
                style=node_style,
                label=f'<{rendered_node}>',
                fontname='Roboto',
                margin=0.2,
                penwidth=pen_width,
                **href_options,
                **self._node_options(row),
            )

        node_iris = set(map(
            operator.itemgetter('node'),
            node_rows,
        ))

        edges_rows = list(
            self.remove_duplicate_edges(
                self.query(self.edges_query),
            ),
        )

        for row in edges_rows:
            if row['source'] in node_iris and row['destination'] in node_iris:
                direction = 'forward'
                edge_color = '#00000052'
                if row['edge'] in {
                    IBIS.term('supported-by'),
                    IBIS.term('opposed-by'),
                    IBIS.term('questioned-by'),
                }:
                    direction = 'back'

                graph.add_edge(
                    row['source'],
                    row['destination'],
                    fontname='Roboto',

                    penwidth=2,
                    dir=direction,
                    color=edge_color,

                    **dict(self._edge_options(row)),
                )

        return graph.draw(format='svg', prog='dot').decode('utf-8')

    def remove_duplicate_edges(self, rows):
        """
        ibis:questioned-by is an rdfs:subPropertyOf of ibis:suggests.

        This creates duplicates. They must be removed.
        """
        questioned_by_edges = {
            (row['source'], row['destination'])
            for row in rows
            if row['edge'] == IBIS.term('questioned-by')
        }

        for row in rows:
            if (
                row['edge'] == IBIS.suggests
                and (row['source'], row['destination']) in questioned_by_edges
            ):
                continue

            yield row



from typing import Iterator, Optional

import rdflib
from mkdocs.structure.pages import Page
from octadocs.query import query
from octadocs.types import LOCAL, Quad, SelectResult, Triple


def iri_by_page(page: Page) -> rdflib.URIRef:
    """Convert src_path of a file to a Zet IRI."""
    return src_path_to_iri(page.file.src_path)


def triples_to_quads(
    triples: Iterator[Triple],
    graph: rdflib.URIRef,
) -> Iterator[Quad]:
    """Convert sequence of triples to sequence of quads."""
    yield from (
        triple.as_quad(graph)
        for triple in triples
    )


def src_path_to_iri(src_path: str) -> rdflib.URIRef:
    """Convert src_path of a file to a Zet IRI."""
    return rdflib.URIRef(f'{LOCAL}{src_path}')



import logging
from dataclasses import dataclass
from typing import Any, Dict, List

from octadocs.pyld_document_loaders.base import DocumentLoader
from octadocs.pyld_document_loaders.models import JsonLDDocument
from pyld.jsonld import JsonLdError

logger = logging.getLogger(__name__)


@dataclass
class ChoiceLoader(DocumentLoader):
    """Choose the first loader in the list which works."""

    loaders: List[DocumentLoader]

    def __call__(self, url: str, options: Dict[str, Any]) -> JsonLDDocument:
        """Try each loader in the sequence."""
        if not self.loaders:
            raise JsonLdError(
                message=(
                    'Please specify at least one loader '
                    'for ChoiceLoader to choose from.'
                ),
                type_='jsonld.InvalidUrl',
                details={'url': url},
                code='loading document failed',
            )

        errors = []
        for loader in self.loaders:
            try:
                return loader(url=url, options=options)
            except JsonLdError as err:
                errors.append(err)

        raise JsonLdError(
            message=f'None of the loaders could process URL: {url}',
            type_='jsonld.InvalidUrl',
            details={
                'url': url,
                'errors': errors,
            },
            code='loading document failed',
        )



import itertools
import json
from dataclasses import dataclass, replace
from typing import Any, Iterable, Optional, TextIO

from documented import DocumentedError
from iolanta.loaders.base import Loader
from iolanta.models import LDContext, LDDocument, Quad
from iolanta.namespaces import LOCAL
from iolanta.parsers.base import Parser
from octadocs.types import OCTA
from pyld import jsonld
from pyld.jsonld import JsonLdError
from rdflib import XSD, URIRef, RDF
from rdflib.term import BNode, Literal, Node


def assign_key_if_not_present(  # type: ignore
    document: LDDocument,
    key: str,
    default_value: Any,
) -> LDDocument:
    """Add key to document if it does not exist yet."""
    if isinstance(document, dict):
        if document.get(key) is None:
            return {
                key: default_value,
                **document,
            }

        return document

    elif isinstance(document, list):
        return [
            assign_key_if_not_present(
                document=sub_document,
                key=key,
                default_value=default_value,
            )
            for sub_document in document
        ]

    return document


@dataclass
class UnresolvedIRI(DocumentedError):
    """
    An unresolved IRI found.

        IRI: {self.iri}
        file: {self.file}
        prefix: {self.prefix}

    Perhaps you forgot to import appropriate context? For example:

    ```yaml
    $context:
        - $import: {self.prefix}
    ```
    """

    iri: str
    prefix: str
    file: Optional[str] = None


def raise_if_term_is_qname(term_value: str):
    """Raise an error if a QName is provided instead of a full IRI."""
    prefix, etc = term_value.split(':', 1)

    if etc.startswith('/'):
        return

    if prefix in {'local', 'templates'}:
        return

    raise UnresolvedIRI(
        iri=term_value,
        prefix=prefix,
    )


def parse_term(
    term,
    blank_node_prefix,
) -> Node:
    """Parse N-Quads term into a Quad."""
    term_type = term['type']
    term_value = term['value']

    if term_type == 'IRI':
        raise_if_term_is_qname(term_value)
        return URIRef(term_value)

    if term_type == 'literal':
        language = term.get('language')

        if datatype := term.get('datatype'):
            datatype = URIRef(datatype)

            # XSD.string does not provide any extra information. Removing it.
            if datatype == XSD.string:
                datatype = None

        if language and datatype:
            datatype = None

        return Literal(
            term_value,
            datatype=datatype,
            lang=language,
        )

    if term_type == 'blank node':
        return BNode(
            value=term_value.replace('_:', f'{blank_node_prefix}/'),
        )

    raise ValueError(f'Unknown term: {term}')


def parse_quads(
    quads_document,
    graph: URIRef,
    blank_node_prefix: str = '',
) -> Iterable[Quad]:
    """Parse an N-Quads output into a Quads stream."""
    for graph_name, quads in quads_document.items():
        if graph_name == '@default':
            graph_name = graph

        else:
            graph_name = URIRef(graph_name)

        for quad in quads:
            try:
                yield Quad(
                    subject=parse_term(quad['subject'], blank_node_prefix),
                    predicate=parse_term(quad['predicate'], blank_node_prefix),
                    object=parse_term(quad['object'], blank_node_prefix),
                    graph=graph_name,
                )
            except UnresolvedIRI as unresolved_iri:
                raise replace(
                    unresolved_iri,
                    file=str(graph),
                )


class JSON(Parser):
    """Load JSON data."""

    def as_jsonld_document(self, raw_data: TextIO) -> LDDocument:
        """Read JSON content as a JSON-LD document."""
        return json.load(raw_data)

    def as_quad_stream(
        self,
        raw_data: TextIO,
        iri: Optional[URIRef],
        context: LDContext,
        root_loader: Loader,
    ) -> Iterable[Quad]:
        """Read JSON-LD data into a quad stream."""
        document = self.as_jsonld_document(raw_data)

        document = assign_key_if_not_present(
            document=document,
            key='@id',
            default_value=str(iri),
        )

        document = assign_key_if_not_present(
            document=document,
            key=str(OCTA.subjectOf),
            default_value={
                '@id': str(iri),
            },
        )

        try:
            document = jsonld.expand(
                document,
                options={
                    'expandContext': context,
                    'documentLoader': root_loader,

                    # Explanation:
                    #   https://github.com/digitalbazaar/pyld/issues/143
                    'base': str(LOCAL),
                },
            )
        except JsonLdError as err:
            raise ExpandError(
                message=str(err),
                document=document,
                context=context,
                iri=iri,
            ) from err

        document = jsonld.flatten(document)

        static_quads = [
            Quad(iri, RDF.type, OCTA.File, iri),
        ]

        return list(
            itertools.chain(
                parse_quads(
                    quads_document=jsonld.to_rdf(document),
                    graph=iri,
                    blank_node_prefix=str(iri),
                ),
                static_quads,
            )
        )


@dataclass
class ExpandError(DocumentedError):
    """
    JSON-LD expand operation failed.

    IRI: {self.iri}
    {self.message}
    """

    message: str
    document: LDDocument
    context: LDContext
    iri: str

    @property
    def formatted_data(self) -> str:
        """Format document for printing."""
        return json.dumps(self.document, indent=2)

    @property
    def formatted_context(self):
        """Format context for printing."""
        return json.dumps(self.context, indent=2)



from dominate.tags import a, span  # noqa: WPS347
from iolanta.facet import Facet
from more_itertools import first
from rdflib.term import Literal, URIRef


def language_by_row(row):
    try:
        return row['label'].language
    except KeyError:
        return None


class Default(Facet):
    """Default facet for rendering an unknown node."""

    sparql_query = '''
    SELECT * WHERE {
        OPTIONAL {
            ?page rdfs:label ?label .
        }

        OPTIONAL {
            ?page octa:symbol ?symbol .
        }

        OPTIONAL {
            ?page octa:url ?url .
        }

        OPTIONAL {
            ?page rdfs:comment ?comment .
        }

        OPTIONAL {
            ?page a octa:Page .
            BIND(true AS ?is_page)
        }
    } ORDER BY ?label
    '''

    def retrieve_row(self, page: URIRef):
        rows = self.query(
            self.sparql_query,
            page=page,
        )

        row_by_language = {
            language_by_row(row): row
            for row in rows
        }

        return row_by_language.get(
            self.language,
        ) or row_by_language.get(None)

    def render(self):   # noqa: C901
        """Render HTML."""
        if isinstance(self.iri, Literal):
            return str(self.iri.value)

        description = self.retrieve_row(self.uriref)
        if description is None:
            return str(self.iri)

        label = description.get('label', str(self.iri))
        url = description.get('url')

        symbol = description.get('symbol')
        if not symbol:
            symbol = ''

        comment = description.get('comment')

        if url:
            return a(
                symbol,
                label,
                href=url,
                title=comment,
            )

        if comment:
            return span(label, title=comment)

        return label



import logging
import operator
import re
from functools import lru_cache, partial, cached_property
from pathlib import Path
from typing import Any, Callable, Dict, Optional

import rdflib
from mkdocs.livereload import LiveReloadServer
from mkdocs.plugins import BasePlugin
from mkdocs.structure.files import Files
from mkdocs.structure.nav import Navigation
from mkdocs.structure.pages import Page
from urlpath import URL

from iolanta import as_document
from iolanta.models import LDDocument
from octadocs.conversions import iri_by_page, src_path_to_iri
from octadocs.default_context import construct_root_context
from octadocs.iolanta import render
from octadocs.iolanta.errors import FacetNotFound
from octadocs.iolanta.iolanta import Render
from octadocs.macros import link
from octadocs.language_from_config import language_from_config
from octadocs.navigation.processor import OctadocsNavigationProcessor
from octadocs.octiron import Octiron
from octadocs.query import query
from octadocs.storage import save_graph
from octadocs.stored_query import StoredQuery
from octadocs.types import DEFAULT_NAMESPACES, LOCAL, OCTA, Query
from rdflib import URIRef
from rich.traceback import install
from typing_extensions import TypedDict

# install(show_locals=False)

logger = logging.getLogger(__name__)


class ConfigExtra(TypedDict):
    """Extra portion of the config which we put our graph into."""

    graph: rdflib.ConjunctiveGraph
    octiron: Octiron
    queries: StoredQuery
    named_contexts: Dict[str, Any]


class Config(TypedDict):
    """MkDocs configuration."""

    docs_dir: str
    extra: ConfigExtra
    nav: dict   # type: ignore


class TemplateContext(TypedDict):
    """Context for the native MkDocs page rendering engine."""

    graph: rdflib.ConjunctiveGraph
    iri: rdflib.URIRef
    this: rdflib.URIRef
    query: Query
    queries: StoredQuery
    local: rdflib.Namespace
    render: Callable[[rdflib.URIRef], str]

    # FIXME this is hardcode and should be removed
    rdfs: rdflib.Namespace


@lru_cache(None)
def cached_octiron(docs_dir: Path, language: Optional[str]) -> Octiron:
    """Retrieve cached Octiron instance or create it if absent."""
    return Octiron(
        root_directory=docs_dir,
        root_context=construct_root_context(
            namespaces=DEFAULT_NAMESPACES,
        ),
        language=language,
    )


class OctaDocsPlugin(BasePlugin):
    """MkDocs Meta plugin."""

    octiron: Octiron
    stored_query: StoredQuery

    def retrieve_named_context(self, relative_path: str) -> LDDocument:
        """Construct URL to the named context file."""
        return as_document(
            URL(
                'file://',
                Path(__file__).parent / 'data' / relative_path,
            ),
        )

    def on_config(self, config: Config) -> Config:
        """Initialize Octiron and provide graph to macros through the config."""
        docs_dir = Path(config['docs_dir'])

        self.octiron = cached_octiron(
            docs_dir=docs_dir,
            language=language_from_config(config),
        )

        self.stored_query = StoredQuery(
            path=docs_dir.parent / 'queries',
            executor=partial(
                query,
                instance=self.octiron.graph,
            ),
        )

        if config['extra'] is None:
            config['extra'] = {}  # type: ignore

        config['extra'].update({
            'graph': self.octiron.graph,
            'octiron': self.octiron,
            'queries': self.stored_query,

            'named_contexts': {
                'rdfs': self.retrieve_named_context('rdfs/named-context.json'),
                'iolanta': self.retrieve_named_context(
                    'iolanta/named-context.json',
                ),
                'owl': self.retrieve_named_context('owl/named-context.json'),
                # 'octa': self.retrieve_named_context('octa/named-context.json'),
            },
        })

        return config

    def on_files(self, files: Files, config: Config):
        """Extract metadata from files and compose the site graph."""
        # Load the Octadocs vocabulary into graph
        self.octiron.update_from_file(
            path=Path(__file__).parent / 'yaml/octadocs.yaml',
            local_iri=rdflib.URIRef(OCTA),
            global_url='/octadocs.yaml',
            named_contexts=config['extra']['named_contexts'],
        )

        # And the global iolanta vocabulary
        self.octiron.update_from_file(
            path=Path(__file__).parent / 'yaml/iolanta.yaml',
            local_iri=rdflib.URIRef('https://iolanta.tech/'),
            global_url='/iolanta.yaml',
            named_contexts=config['extra']['named_contexts'],
        )

        for mkdocs_file in files:
            self.octiron.update_from_file(
                path=Path(mkdocs_file.abs_src_path),
                local_iri=src_path_to_iri(mkdocs_file.src_path),
                global_url=f'/{mkdocs_file.url}',
                named_contexts=config['extra']['named_contexts'],
            )

    def on_page_markdown(
        self,
        markdown: str,
        page: Page,
        config: Config,
        files: Files,
    ):
        """Inject page template path, if necessary."""
        page.iri = iri_by_page(page)
        page.id = self.id_by_page.get(page.iri, page.iri)

        try:
            template_url = Render(
                ldflex=self.octiron.ldflex,
            ).find_facet_iri(
                node=page.id,
                environments=[URIRef(OCTA)],
            )
        except FacetNotFound:
            return markdown

        page.meta['template'] = re.sub(
            '^templates:/*',
            '',
            template_url,
        )

        return markdown

    @cached_property
    def id_by_page(self) -> Dict[URIRef, URIRef]:
        """Retrieve the best suitable IRI for an MkDocs page."""
        rows = self.octiron.ldflex.query(
            'SELECT ?page ?this WHERE { ?this octa:subjectOf ?page . }',
        )

        return {
            row['page']: row['this']
            for row in rows
        }

    def on_page_context(
        self,
        context: TemplateContext,
        page: Page,
        config: Config,
        nav: Page,
    ) -> TemplateContext:
        """Attach the views to certain pages."""
        page_iri = rdflib.URIRef(
            f'{LOCAL}{page.file.src_path}',
        )

        context['this'] = context['id'] = self.id_by_page.get(page_iri, page_iri)
        context['graph'] = self.octiron.graph
        context['iri'] = page_iri

        # noinspection PyTypedDict
        context['query'] = partial(
            query,
            instance=self.octiron.graph,
        )
        context['queries'] = self.stored_query

        context['local'] = LOCAL
        context['LOCAL'] = LOCAL
        context.update(self.octiron.namespaces)

        context['render'] = partial(
            render,
            octiron=self.octiron,
        )

        # Provide all the support namespaces into template context
        context['octiron'] = self.octiron
        context['link'] = partial(
            link,
            octiron=self.octiron,
        )

        # Page attributes
        page.iri = page_iri
        page.id = self.id_by_page.get(page_iri, page_iri)

        return context

    def on_nav(
        self,
        nav: Navigation,
        config: Config,
        files: Files,
    ) -> Navigation:
        """Update the site's navigation from the knowledge graph."""
        # This must run after all on_files() handlers of all plugins, but before
        # any page rendering and facets. Nav calculation depends on inference,
        # that's why we call it here in on_nav().
        self.octiron.apply_inference()

        if not config.get('nav'):
            nav = OctadocsNavigationProcessor(
                graph=self.octiron.graph,
                navigation=nav,
            ).generate()

        return nav

    def on_serve(
        self,
        server: LiveReloadServer,
        config: Config,
        builder,
    ) -> LiveReloadServer:
        inference_directory = Path(config['docs_dir']).parent / 'inference'
        if inference_directory.is_dir():
            for sparql_file in inference_directory.glob('**/*.sparql'):
                server.watch(str(sparql_file), builder)

        queries_directory = Path(config['docs_dir']).parent / 'queries'
        if queries_directory.is_dir():
            for sparql_file in queries_directory.glob('**/*.sparql'):
                server.watch(str(sparql_file), builder)

        return server

    def on_post_build(self, config: Config):
        """Save the graph to a file on disk."""
        self.octiron.save_graph_to_cache()

    def on_build_error(self, error):
        """Save the graph to a file on build error."""
        self.octiron.save_graph_to_cache()



from dataclasses import dataclass
from typing import Any

from documented import DocumentedError
from rdflib import Namespace

ADR = Namespace('https://adr.octadocs.io/')


@dataclass
class ADRNumberInvalid(DocumentedError):
    """
    The value `{self.number}` is not a valid ADR number.

    The number must be natural (i. e. a positive integer).

    ADR page to blame:

        {self.page}
    """

    number: Any
    page: str



import json
from pathlib import Path

import rich
from typer import Typer, Option
from urlpath import URL

from iolanta.loaders import LocalFile
from iolanta.shortcuts import construct_root_loader
from ldflex import LDFlex
from octadocs.cli.formatters.csv import csv_print
from octadocs.cli.formatters.json import print_json
from octadocs.cli.formatters.pretty import pretty_print
from octadocs.conversions import src_path_to_iri
from octadocs.default_context import construct_root_context
from octadocs.storage import load_graph
from octadocs.types import QueryResultsFormat

app = Typer(name='show')


def find_docs_dir() -> Path:
    """Find the docs dir of the MkDocs site."""
    cwd = Path.cwd()

    while True:
        docs = cwd / 'docs'
        if docs.is_dir():
            return docs

        cwd = cwd.parent


@app.command(name='file')
def show_file(
    path: Path,
    fmt: QueryResultsFormat = Option(
        default=QueryResultsFormat.PRETTY,
        metavar='format',
    ),
):
    """Show graph from a file."""
    docs_dir = find_docs_dir()
    path = path.absolute().relative_to(docs_dir)

    iri = src_path_to_iri(str(path))

    graph = load_graph(Path.cwd() / '.cache/octadocs')
    ldflex = LDFlex(graph)

    query_result = ldflex.query(
        '''
        SELECT ?s ?p ?o WHERE {
            GRAPH ?g {
                ?s ?p ?o
            }
        }
        ''',
        g=iri,
    )

    {
        QueryResultsFormat.CSV: csv_print,
        QueryResultsFormat.PRETTY: pretty_print,
        QueryResultsFormat.JSON: print_json,
    }[fmt](query_result)  # type: ignore


@app.command(name='context')
def show_context(path: Path):
    docs_dir = find_docs_dir()

    path = path.absolute().relative_to(docs_dir)
    url = URL(f'file://{path}')

    root_context = construct_root_context({})
    loader = LocalFile(
        root_directory=docs_dir,
        default_context=root_context,
    )

    context = loader.find_context(url)

    rich.print(
        json.dumps(
            context,
            ensure_ascii=False,
            indent=2,
        ),
    )



from mkdocs.config import load_config
from more_itertools import consume
from octadocs.cli import context, sparql, show
from typer import Typer


def generate_app() -> Typer:
    app = Typer()

    app.add_typer(sparql.app)
    app.add_typer(context.app)
    app.add_typer(show.app)

    config = load_config()

    typer_instances = [
        plugin.typer()
        for plugin in config['plugins'].values()
        if hasattr(plugin, 'typer')
    ]

    typer_instances = filter(bool, typer_instances)

    consume(
        map(
            app.add_typer,
            typer_instances,
        ),
    )

    return app


app = generate_app()



from typing import Any, Dict, Optional, TypedDict


class JsonLDDocument(TypedDict):
    """Loaded document."""

    document: Dict[str, Any]
    contextUrl: Optional[str]
    documentUrl: str
    contentType: str



from pathlib import Path
from typing import Dict

from iolanta import as_document
from iolanta.conversions import path_to_url
from iolanta.models import LDDocument
from rdflib import Namespace


def construct_root_context(namespaces: Dict[str, Namespace]) -> LDDocument:
    """
    Construct default JSON-LD context for all Octadocs data.

    This context contains:
        - declarations for all namespaces,
        - plus content of the default `context.yaml`.
    """
    namespaces_context = {
        prefix: str(namespace)
        for prefix, namespace in namespaces.items()
        if prefix
    }
    document_context = as_document(
        path_to_url(
            Path(__file__).parent / 'yaml/context.yaml',
        ),
    )

    return {
        **namespaces_context,
        **document_context,
    }



from octadocs_adr.facets.adr_list import adr_list
from octadocs_adr.facets.sidebar import PageSidebar, sidebar_property
from octadocs_adr.facets.status import status
from octadocs_adr.facets.status_class import status_class



from typing import Any, Dict, List, Union

from iolanta.parsers.json import ExpandError
from octadocs.types import LOCAL
from pyld import jsonld
from pyld.jsonld import JsonLdError

try:  # noqa
    from yaml import CSafeDumper as SafeDumper  # noqa
    from yaml import CSafeLoader as SafeLoader  # noqa
except ImportError:
    from yaml import SafeDumper  # type: ignore   # noqa
    from yaml import SafeLoader  # type: ignore   # noqa

MetaData = Union[List[Dict[str, Any]], Dict[str, Any]]  # type: ignore  # noqa


def jsonld_expand(data, context, local_iri, pyld_loader=None):
    """Expand the JSON_LD data with the help of particular loader."""
    options = {
        'base': str(LOCAL),
    }

    if context:
        options.update({'expandContext': context})

    if pyld_loader is not None:
        options.update({'documentLoader': pyld_loader})

    try:
        return jsonld.expand(
            data,
            options=options,
        )
    except (TypeError, JsonLdError) as err:
        raise ExpandError(
            message=str(err),
            document=data,
            context=context,
            iri=local_iri,
        ) from err



from datetime import date, datetime
from typing import Iterable

from dominate.tags import code, a, span, br, html_tag
from more_itertools import first

from iolanta.facet import Facet


class Release(Facet):
    """Render a software project release."""

    sparql = '''
    PREFIX github: <https://octadocs.io/github/>

    SELECT * WHERE {
        $release github:name ?name .

        OPTIONAL {
            $release github:published_at ?date .
        }

        OPTIONAL {
            $release github:prerelease ?prerelease .
        }
    }
    '''

    def construct_content(self) -> Iterable[html_tag]:
        """Construct release description."""
        rows = self.query(
            self.sparql,
            release=self.iri,
        )

        row = first(rows)

        date_value = row.get('date')

        if date_value is not None:
            date_value = date_value.value

            if isinstance(date_value, datetime):
                date_value = date_value.date()

            yield code(str(date_value))
            yield br()

        yield a(
            row['name'],
            href=row['release'],
        )

    def render(self):
        """Draw release description."""
        return span(
            *self.construct_content(),
            cls='octadocs-github-release',
        )



import logging
from functools import cached_property
from pathlib import Path
from typing import Dict, Any, Set

from funcy import pluck
from mkdocs.plugins import BasePlugin
from mkdocs.structure.files import Files
from mkdocs.structure.nav import Navigation
from mkdocs.structure.pages import Page
from rdflib import URIRef
from urlpath import URL

from iolanta import as_document
from iolanta.models import LDContext
from octadocs.mixins import OctadocsMixin
from octadocs.plugin import Config
from octadocs_ibis.models import IBIS

logger = logging.getLogger(__name__)


POSITIONS_WITH_EVALUATIONS = '''
INSERT {
    ?position octa:symbol ?symbol .
} WHERE {
    ?position a ibis:Position .

    ?position ibis:evaluation / prov:value* ?is_approved .
    FILTER (isLiteral(?is_approved)) .

    BIND(
        IF(?is_approved = TRUE, "✔️", "❌")
        AS ?symbol
    )
}
'''


class IbisPlugin(OctadocsMixin, BasePlugin):
    """Issue-based information system (IBIS) plugin."""

    plugin_data_dir = Path(__file__).parent / 'data'
    namespaces = {
        'ibis': IBIS,
    }

    @cached_property
    def ibis_page_uris(self) -> Set[URIRef]:
        """IRIs of all pages which describe IBIS nodes."""
        return set(
            pluck(
                'page',
                self.query(
                    '''
                    SELECT ?page WHERE {
                        { ?thing a ibis:Issue } UNION { ?thing a ibis:Position }
                        ?thing octa:subjectOf* ?page .
                    }
                    ''',
                ),
            ),
        )

    def named_contexts(self) -> Dict[str, LDContext]:
        """Reusable named contexts."""
        return {
            'ibis': as_document(
                URL('file://', self.plugin_data_dir / 'named-context.json'),
            ),
        }

    def vocabularies(self) -> Dict[URIRef, Path]:
        """Load IBIS ontology."""
        return {
            URIRef(IBIS): self.plugin_data_dir / 'ibis.json',
            IBIS.term('facets.yaml'): self.plugin_data_dir / 'facets.yaml',
        }

    @property
    def templates_path(self) -> Path:
        """Templates associated with the plugin."""
        return Path(__file__).parent / 'templates'

    def on_page_context(
        self,
        context: Dict[str, Any],
        page: Page,
        **kwargs,
    ):
        """Make custom functions available to the template."""
        return {
            **context,
            'ibis': IBIS,
        }

    def on_nav(
        self,
        nav: Navigation,
        config: Config,
        files: Files,
    ) -> Navigation:
        logger.info('Marking ibis:Positions with symbols...')
        self.update(POSITIONS_WITH_EVALUATIONS)
        logger.info('Symbols marked.')

        return nav

    def on_page_markdown(
        self,
        markdown: str,
        page: Page,
        config: Config,
        files: Files,
    ):
        """Inject page template path, if necessary."""
        if page.iri in self.ibis_page_uris:
            page.meta['hide'] = ['navigation']

        return markdown



from octadocs_ibis.facets.arguments_for_position import ArgumentsForPosition
from octadocs_ibis.facets.breadcrumbs import Breadcrumbs
from octadocs_ibis.facets.position_evaluation import PositionEvaluation
from octadocs_ibis.facets.argument import Argument
from octadocs_ibis.facets.positions_for_issue import PositionsForIssue
from octadocs_ibis.facets.suggested_next_steps import SuggestedNextSteps
from octadocs_ibis.facets.sources_for_position import SourcesForPosition
from octadocs_ibis.facets.source_for_position import SourceForPosition
from octadocs_ibis.facets.image import Image
from octadocs_ibis.facets.ibis_graph import IBISGraph
from octadocs_ibis.facets.ibis_node import IBISNode



from boltons.iterutils import chunked_iter
from dominate.tags import table, tr, td, tbody, div

from iolanta.facet import Facet
from iolanta.namespaces import IOLANTA
from octadocs.iolanta import render
from octadocs_ibis.models import IBIS

# Number of columns in the sources table.
COLUMN_COUNT = 2


class SourcesForPosition(Facet):
    """List sources for a given Position."""

    sparql_query = '''
    SELECT ?source WHERE {
        $position prov:wasDerivedFrom ?source .

        OPTIONAL {
            ?source octa:position ?explicit_position .
        }

        BIND(COALESCE(?explicit_position, 0) as ?position)
    } ORDER BY ?position
    '''

    def render(self):
        sources = self.query(
            self.sparql_query,
            position=self.uriref,
        )

        renderables = [render(
            source['source'],
            environments=[IBIS.SourcesForPosition, IOLANTA.html],
            octiron=self.octiron,
        ) for source in sources]

        if renderables:
            return div(*renderables)

        return ''



import json
from dataclasses import dataclass, field
from functools import reduce
from pathlib import Path
from typing import Iterable, List, Optional, TextIO, Type

from documented import DocumentedError
from iolanta.context import merge
from iolanta.conversions import url_to_path
from iolanta.loaders.base import Loader
from iolanta.models import LDContext, LDDocument, Quad
from iolanta.parsers.base import Parser
from iolanta.parsers.json import JSON
from iolanta.parsers.markdown import Markdown
from iolanta.parsers.yaml import YAML
from rdflib import URIRef
from urlpath import URL


@dataclass
class IsAContext(DocumentedError):
    """
    The provided file is a context.

        - Path: {self.path}

    This file is not a piece of data and cannot be loaded into the graph.
    """

    path: URL


@dataclass
class ParserNotFound(DocumentedError):
    """
    Parser not found.

        Path: {self.path}
    """

    path: URL


def merge_contexts(*contexts: LDContext) -> LDContext:
    return reduce(
        merge,
        contexts,
    )


@dataclass(frozen=True)
class LocalFile(Loader):
    """
    Retrieve Linked Data from a file on local disk.

    Requires URL with file:// scheme as input.
    """

    root_directory: Optional[Path] = None
    context_filenames: List[str] = field(default_factory=lambda: [
        'context.yaml',
        'context.json',
    ])
    default_context: Optional[LDContext] = field(default=None, repr=False)

    def choose_parser_class(self, url: URL) -> Type[Parser]:
        """
        Choose parser class based on file extension.

        FIXME this is currently hard coded; need to change to a more extensible
          mechanism.
        """
        if url.suffix == '.yaml':
            return YAML

        elif url.suffix == '.json':
            return JSON

        elif url.suffix == '.md':
            return Markdown

        raise ParserNotFound(path=url)

    def as_quad_stream(
        self,
        url: URL,
        iri: Optional[URIRef],
        root_loader: Loader,
    ) -> Iterable[Quad]:
        """Extract a sequence of quads from a local file."""
        if url.stem == 'context':
            raise IsAContext(path=url)

        parser_class = self.choose_parser_class(url)

        context = self.find_context(url)

        with url_to_path(url).open() as text_io:
            return parser_class().as_quad_stream(
                raw_data=text_io,
                iri=iri,
                context=context,
                root_loader=root_loader,
            )

    def as_file(self, url: URL) -> TextIO:
        """Construct a file-like object."""
        path = url_to_path(url)
        with path.open() as text_io:
            return text_io

    def as_jsonld_document(
        self,
        url: URL,
        iri: Optional[URIRef] = None,
    ) -> LDDocument:
        """As JSON-LD document."""
        parser_class: Type[Parser] = self.choose_parser_class(url)
        with url_to_path(url).open() as text_io:
            document = parser_class().as_jsonld_document(text_io)

        if iri is not None and isinstance(document, dict):
            document.setdefault('@id', str(iri))

        return document

    def find_context(self, url: URL) -> LDContext:
        """Traverse the directories and construct context."""
        return merge_contexts(
            self.default_context,
            *self.contexts_by_url(url),
        )

    def contexts_by_url(self, url: URL) -> Iterable[LDContext]:
        return [
            self.as_jsonld_document(
                url=URL(f'file://{context_file}'),
            )
            for context_file in self.context_files_by_url(url)
        ]

    def context_files_by_url(self, url: URL) -> Iterable[Path]:
        """Yield all contexts by URL."""
        ancestor_directories = self.ancestors_by_url(url)
        for directory in ancestor_directories:
            for filename in self.context_filenames:
                if (context_file := directory / filename).exists():
                    yield context_file

    def ancestors_by_url(self, url: URL) -> Iterable[Path]:
        """Find all ancestor directories to this path."""
        if self.root_directory is None:
            raise ValueError('Please specify a root directory.')

        root_directory = self.root_directory.absolute()
        for ancestor in reversed(url_to_path(url).parents):
            ancestor = ancestor.absolute()

            # Replace with .is_relative_to() for Python 3.9.
            if str(ancestor).startswith(str(root_directory)):
                yield ancestor



from dominate.tags import span, a
from iolanta.facet import Facet
from more_itertools import first


class GithubLicense(Facet):
    """Render license information."""

    sparql_text = '''
    SELECT * WHERE {
        $license
            <http://spdx.org/rdf/terms#licenseId> ?id ;
            <http://spdx.org/rdf/terms#name> ?name .

        OPTIONAL {
            $license rdfs:label ?label .
        }
    } ORDER BY ?id
    '''

    def render(self):
        """Render."""
        rows = self.query(
            self.sparql_text,
            license=self.iri,
        )

        row = first(rows)

        url = row['license']
        title = row['name']
        text = str(row['id'])

        if text == 'NOASSERTION':
            text = title

        label = row.get('label')

        return a(
            label or text,
            title=title,
            href=url,
        )



from dataclasses import dataclass

from documented import DocumentedError
from more_itertools import first
from octadocs.octiron import Octiron
from rdflib import URIRef
from rdflib.term import Node
from urlpath import URL


@dataclass
class StatusNotFound(DocumentedError):
    """`{self.readable_status}` is not a valid ADR document status."""

    status: Node

    @property
    def readable_status(self) -> str:
        """Render status in a readable form."""
        url = URL(str(self.status))

        if url.scheme == 'local':
            return url.path

        return str(url)


def status(octiron: Octiron, iri: URIRef):
    """Render ADR document status."""
    meta = first(
        octiron.query(
            '''
            SELECT * WHERE {
                ?status rdfs:label ?label .
                
                OPTIONAL {
                    ?status octa:symbol ?symbol .
                }
            } ORDER BY DESC(?symbol)
            ''',
            status=iri,
        ),
        None,
    )

    if not meta:
        raise StatusNotFound(status=iri)

    label = meta['label']
    if symbol := meta.get('symbol'):
        label = f'{symbol} {label}'

    return label



import itertools
import operator
from typing import Iterable, List

from dominate.tags import table, tr, td, div, h2, h3
from more_itertools import grouper
from rdflib import URIRef

from iolanta.facet import Facet
from iolanta.namespaces import IOLANTA
from octadocs.iolanta import render
from octadocs_ibis.models import IBIS
from octadocs_table.models import TABLE

COLUMN_COUNT = 3


class ArgumentsForPosition(Facet):
    """Render arguments list of a given Position."""

    query_text = '''
    SELECT ?arg ?link WHERE {
        ?arg
            ?link $position ;
            a ibis:Argument .
    } ORDER BY ?link ?arg
    '''

    def construct_table_rows(self) -> Iterable[tr]:
        rows = self.query(
            self.query_text,
            position=self.iri,
        )

        rows_by_relation = itertools.groupby(
            rows,
            key=operator.itemgetter('link'),
        )

        return {
            relation: [item['arg'] for item in items]
            for relation, items in rows_by_relation
        }

    def format_table_rows(self, arguments: List[URIRef]) -> Iterable[tr]:
        """Format table rows for a gallery of arguments."""
        env = [TABLE.td, IOLANTA.html]
        renderables = [render(
            argument,
            environments=env,
            octiron=self.octiron,
        ) for argument in arguments]

        grouped_renderables = grouper(renderables, COLUMN_COUNT, '')

        for group in grouped_renderables:
            cells = [td(renderable) for renderable in group]
            yield tr(*cells)

    def render(self):
        arguments_by_relation = self.construct_table_rows()

        if not arguments_by_relation:
            return ''

        with div() as container:
            h2('Аргументы')

            for relation, arguments in arguments_by_relation.items():
                h3(render(
                    relation,
                    environments=[IOLANTA.td, IOLANTA.html],
                    octiron=self.octiron,
                ))

                table_rows = self.format_table_rows(arguments)

                table(*table_rows)

        return container



from rdflib import Literal


class LanguageAware:
    """Language aware facet."""

    def _is_row_fitting_language(self, row) -> bool:
        """All literals in the row either have no language or match the site."""
        for field_name, field_value in row.items():
            if (
                isinstance(field_value, Literal)
                and field_value.language
                and field_value.language != self.octiron.language
            ):
                return False

        return True

    def filter_by_language(self, rows):
        return filter(
            self._is_row_fitting_language,
            rows,
        )

    def query(self, query_text: str, **kwargs):
        """Language aware query."""
        return self.filter_by_language(
            super().query(query_text, **kwargs),
        )



from pathlib import Path
from typing import Dict

from iolanta import as_document
from iolanta.models import LDContext
from mkdocs.plugins import BasePlugin
from octadocs.mixins import OctadocsMixin
from octadocs.types import OCTA
from rdflib import PROV, URIRef
from urlpath import URL


class ProvenancePlugin(OctadocsMixin, BasePlugin):
    """Render an HTML table from data presented in the graph."""

    plugin_data_dir = Path(__file__).parent / 'data'

    def named_contexts(self) -> Dict[str, LDContext]:
        """Reusable named contexts."""
        return {
            'prov': as_document(
                URL('file://', self.plugin_data_dir / 'named-context.json'),
            ),
        }

    def vocabularies(self) -> Dict[URIRef, Path]:
        """Load PROV-O ontology."""
        return {
            URIRef(PROV): self.plugin_data_dir / 'prov.json',
            URIRef(OCTA.term('prov')): (
                self.plugin_data_dir / 'inference.yaml'
            ),
        }



"""
Iolanta facet management.

This module contains a few functions which later will be refactored into
Iolanta - the generic metaverse/cyberspace browser.
"""
import pydoc
import random
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, List, Optional, Union, Type

from dominate.tags import (
    div, small, code, br, span, strong, b, font, table,
    tr, td,
)
from dominate.util import raw
from rdflib.term import Literal, Node, URIRef

from iolanta.facet import Facet
from iolanta.facets.base import FacetSearchAttempt
from iolanta.facets.by_environment import FindFacetByEnvironment
from iolanta.facets.by_instance import FindFacetByInstance
from iolanta.facets.by_literal_datatype import FindFacetByLiteralDatatype
from iolanta.facets.by_type import FindFacetByInstanceTypes
from iolanta.facets.for_literal import FindFacetForLiteral
from ldflex import LDFlex
from octadocs.iolanta.errors import FacetError, FacetNotCallable, FacetNotFound
from octadocs.octiron import Octiron

HTML = URIRef('https://html.spec.whatwg.org/')


def resolve_facet(iri: URIRef) -> Type[Facet]:
    """Resolve a path to a Python object to that object."""
    url = str(iri)

    if not url.startswith('python://'):
        raise Exception(
            'Octadocs only supports facets which are importable Python '
            'callables. The URLs of such facets must start with `python://`, '
            'which {url} does not comply to.'.format(
                url=url,
            )
        )

    # It is impossible to use `urlpath` for this operation because it (or,
    # rather, one of upper classes from `urllib` that `urlpath` depends upon)
    # will lowercase the URL when parsing it - which means, irreversibly. We
    # have to resort to plain string manipulation.
    import_path = url.replace('python://', '').strip('/')

    facet = pydoc.locate(import_path)

    if not callable(facet):
        raise FacetNotCallable(
            path=import_path,
            facet=facet,
        )

    return facet


def _render_facet(
    node: Node,
    facet: Facet,
    environments: Optional[List[URIRef]] = None,
    debug_mode: bool = False,
):
    if debug_mode:
        return debug_node(
            node=node,
            environments=environments,
        )

    return facet.render()


def debug_node(node: Union[str, Node], environments: Optional[List[URIRef]]):
    return table(
        tr(
            td(
                b(font('Debug', color='red')),
                colspan=2,
                border=0,
            ),
        ),
        tr(
            td(b(font('Node', color='red')), border=0, align='left'),
            td(str(node), border=0, align='left'),
        ),
        tr(
            td(b(font('Environments', color='red')), border=0, align='left'),
            td(', '.join(
                map(str, environments),
            ), border=0, align='left'),
        ),
        border=1,
    )


def render(
    node: Union[str, Node],
    octiron: Octiron,
    environments: Optional[List[URIRef]] = None,
) -> str:
    """Find an Iolanta facet for a node and render it."""
    if isinstance(environments, str):
        environments = [environments]

    if not environments:
        environments = [HTML]

    facet_search_attempt = Render(
        ldflex=octiron.ldflex,
    ).search_for_facet(
        node=node,
        environments=environments,
    )

    facet_class = resolve_facet(
        iri=facet_search_attempt.facet,
    )

    if not issubclass(facet_class, Facet):
        raise ValueError(f'{facet_class} is not a Facet.')

    facet = facet_class(
        iri=node,
        octiron=octiron,
        environment=facet_search_attempt.environment,
    )

    try:
        return _render_facet(
            node=node,
            facet=facet,
            environments=environments,
            debug_mode=octiron.is_debug_mode(node),
        )

    except (FacetError, FacetNotFound):   # noqa: WPS329
        # Prevent nested `FacetError` exceptions.
        raise

    except Exception as err:
        raise FacetError(
            node=node,
            facet_iri=facet_search_attempt.facet,
            facet_search_attempt=facet_search_attempt,
            error=err,
        ) from err


@dataclass
class Render:
    """Facet renderer."""

    ldflex: LDFlex

    def search_for_facet(
        self,
        node: Node,
        environments: List[URIRef],
    ) -> FacetSearchAttempt:
        """Find facet IRI for given node for all environments given."""
        facet_search_attempts = list(
            self.attempt_search_for_facet(
                node=node,
                environments=environments,
            ),
        )

        *failures, success = facet_search_attempts

        if success:
            return success

        raise FacetNotFound(
            node=node,
            environments=environments,
            facet_search_attempts=facet_search_attempts,
        )

    def find_facet_iri(
        self,
        node: Node,
        environments: List[URIRef],
    ) -> URIRef:
        """Find facet IRI for given node for all environments given."""
        return self.search_for_facet(
            node=node,
            environments=environments,
        ).facet

    def attempt_search_for_facet(
        self,
        node: Node,
        environments: List[URIRef],
    ) -> Iterable[FacetSearchAttempt]:
        """
        Stream of attempts to find a facet.

        If this function yields a sequence of N elements, first (N - 1) of those
        are guaranteed to be failures. The remaining one element can be both
        a failure (if nothing was found) or a success (if a facet was found).
        """
        for environment in environments:
            facet_searches = self.find_facet_iri_per_environment(
                node=node,
                environment=environment,
            )

            for facet_search in facet_searches:
                yield facet_search

                if facet_search:
                    return

    def find_facet_iri_per_environment(
        self,
        node: Node,
        environment: URIRef,
    ):
        """Find facet IRI for given node in given env."""
        if isinstance(node, Literal):
            yield from self.find_facet_iri_for_literal(
                literal=node,
                environment=environment,
            )

        yield FindFacetByInstance(
            ldflex=self.ldflex,
            node=node,
            environment=environment,
        )

        yield FindFacetByInstanceTypes(
            ldflex=self.ldflex,
            node=node,
            environment=environment,
        )

        yield FindFacetByEnvironment(
            ldflex=self.ldflex,
            node=node,
            environment=environment,
        )

    def __call__(
        self,
        node: Node,
        environment: URIRef,
    ) -> str:
        ...

    def find_facet_iri_for_literal(self, literal: Literal, environment: URIRef):
        """Find facet IRI for a literal."""
        if literal.datatype is not None:
            yield FindFacetByLiteralDatatype(
                ldflex=self.ldflex,
                node=literal,
                environment=environment,
            )

        yield FindFacetForLiteral(
            ldflex=self.ldflex,
            node=literal,
            environment=environment,
        )

        yield FindFacetByEnvironment(
            ldflex=self.ldflex,
            node=literal,
            environment=environment,
        )



from typing import Any, Dict

from octadocs.pyld_document_loaders.models import JsonLDDocument


class DocumentLoader:
    """Load a document by URL."""

    def __call__(self, url: str, options: Dict[str, Any]) -> JsonLDDocument:
        """Execute the loading itself."""
        raise NotImplementedError()
